/* System library. */
#include "sys_defs.h"

#ifdef HAS_MYSQL
#ifdef USE_OPTIMIZERS
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include <mysql.h>

/* Utility library. */
#include "dict.h"
#include "msg.h"
#include "mymalloc.h"
#include "dict_mysql.h"
#include "argv.h"
#include "vstring.h"
#include "split_at.h"
#include "find_inet.h"
#include "optm.h"

#define RETRY_CONN_INTV 60		/* 1 minute */

enum {
  STAT_OK,
  STAT_OPEN,
  STAT_BAD
};

/* need some structs to help organize things */
struct sqlhost_t {
    MYSQL  *db;
    char   *hostname;
    int     stat;			/* STATUNTRIED | STATFAIL | STATCUR */
    time_t  ts;				/* used for attempting reconnection
					 * every so often if a host is down */
};

struct plsql_t {
    int     len_hosts;			/* number of hosts */
    struct sqlhost_t *db_hosts;			
};

struct mysql_t {
  struct sqlconf_t * sqlconf;
  struct plsql_t * plsql;
};

struct  sqlconf_t {
    char   *username;
    char   *password;
    char   *dbname;
    char   *table;
    char   *select_field;
    char   *where_field;
    char  **hostnames;
    int     len_hosts;
}; 

typedef struct sqlhost_t sqlhost_t;
typedef struct sqlconf_t sqlconf_t;
typedef struct plsql_t plsql_t;
typedef struct mysql_t mysql_t;

/* Function prototypes */
mysql_t * mysql_open(const char *);
void mysql_localclose(mysql_t *);
int mysql_add(mysql_t *, char *, int);
int mysql_remove(mysql_t *, int);

/******************************************************************************/
/******************************************************************************/

static sqlconf_t * sqlconf_parse(const char *mysqlcfgfile)
{
    VSTRING * dict_name;
    ARGV   * hosts_argv;
    sqlconf_t * toret = (sqlconf_t *)(mymalloc(sizeof(sqlconf_t)));
    char    * value;

    int     i;

    if(toret == NULL) 
      msg_fatal("optm_mysql, sqlconf_parse: out of memory");

    dict_name = vstring_alloc(64);
    vstring_sprintf(dict_name, "mysql opt dict %s", mysqlcfgfile);
    dict_load_file(vstring_str(dict_name), mysqlcfgfile);

    /* mysql username lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "user"));
    if (value == NULL)
      toret->username = mystrdup("");
    else
      toret->username = mystrdup(value);

    /* password lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "password"));
    if (value == NULL)
      toret->password = mystrdup("");
    else
      toret->password = mystrdup(value);

    /* database name lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "dbname"));
    if (value == NULL)
      msg_fatal("%s: mysql options file does not include database name", mysqlcfgfile);
    else
      toret->dbname = mystrdup(value);

    /* table lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "table"));
    if (value == NULL)
      msg_fatal("%s: mysql options file does not include table name", mysqlcfgfile);
    else
      toret->table = mystrdup(value);

    /* select field lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "select_field"));
    if (value == NULL)
      msg_fatal("%s: mysql options file does not include select field", mysqlcfgfile);
    else
      toret->select_field = mystrdup(value);

    /* where field lookup */
    value=(char *)(dict_lookup(vstring_str(dict_name), "where_field"));
    if (value == NULL)
      msg_fatal("%s: mysql options file does not include where field", mysqlcfgfile);
    else
      toret->where_field = mystrdup(value);

    /* mysql server hosts */
    value=(char *)(dict_lookup(vstring_str(dict_name), "hosts"));
    if (value == NULL)
      value = mystrdup("");
    else
      value = mystrdup(value);
    /* coo argv interface */
    hosts_argv = argv_split(value, " ,\t\r\n");

    if (hosts_argv->argc == 0) {
	argv_add(hosts_argv, "localhost", ARGV_END);
	argv_terminate(hosts_argv);
    }

    toret->len_hosts = hosts_argv->argc;
    toret->hostnames = (char **)(mymalloc(sizeof(char *) * toret->len_hosts));
    if(toret->hostnames == NULL)
      msg_fatal("optm_mysql, sqlconf_parse: out of memory");

    for (i = 0; hosts_argv->argv[i] != NULL; i++) 
      toret->hostnames[i] = mystrdup(hosts_argv->argv[i]);

    myfree(value);
    vstring_free(dict_name);
    argv_free(hosts_argv);
    return toret;
}

/* host_init - initialize HOST structure */
static sqlhost_t host_init(char *hostname)
{
    sqlhost_t host;

    host.stat = STAT_OK;
    host.hostname = mystrdup(hostname);
    host.db = 0;
    host.ts = 0;

    return host;
}

static plsql_t *plsql_init(char *hostnames[], int len_hosts)
{
    plsql_t *toret;
    int     i;

    toret = (plsql_t *)(mymalloc(sizeof(plsql_t)));
    if (toret == NULL)
	msg_fatal("optm_mysql, plsql_init: out of memory");

    toret->len_hosts = len_hosts;
    toret->db_hosts= (sqlhost_t *)(mymalloc(sizeof(sqlhost_t) * len_hosts));
    if (toret->db_hosts == NULL)
	msg_fatal("optm_mysql, plsql_init: out of memory");

    for (i = 0; i < len_hosts; i++) 
	toret->db_hosts[i] = host_init(hostnames[i]);

    return toret;
}

/*****
 * mysql_open */
mysql_t * mysql_open(const char *name)
{
    mysql_t * toret;

    toret = (mysql_t *)(mymalloc(sizeof(mysql_t)));
    if(toret == NULL)
      msg_fatal("optm_mysql, mysql_open: out of memory");

    toret->sqlconf = sqlconf_parse(name);
    toret->plsql = plsql_init(toret->sqlconf->hostnames,
				    toret->sqlconf->len_hosts);
    
    return toret;
}

/******************************************************************************/

static void plmysql_close(sqlhost_t *host)
{
  if(host == NULL)
    return;

  if(host->db) {
    mysql_close(host->db);
    host->db=NULL;
  }
}

static void plmysql_down(sqlhost_t *host)
{
  if (host->stat != STAT_BAD) {
    host->ts = time((time_t *) 0) + RETRY_CONN_INTV;
    host->stat = STAT_BAD;
  }
  plmysql_close(host);
}

static void plmysql_connect(sqlhost_t *host, sqlconf_t *conf)
{
  char   * destination = host->hostname;
  char   * unix_socket = NULL;
  char   * hostname = NULL;
  char   * service;
  unsigned port = 0;

  /*
   * Ad-hoc parsing code. Expect "unix:pathname" or "inet:host:port", where
   * both "inet:" and ":port" are optional.
   */
  if (strncmp(destination, "unix:", 5) == 0) 
    unix_socket = destination + 5;
  else {
    if (strncmp(destination, "inet:", 5) == 0)
      destination += 5;
    hostname = mystrdup(destination);
    if ((service = split_at(hostname, ':')) != 0)
      port = ntohs(find_inet_port(service, "tcp"));
  }

  host->db = mysql_init(NULL);
  if (mysql_real_connect(host->db, hostname, conf->username, conf->password, conf->dbname, port, unix_socket, 0)) {
    host->stat = STAT_OPEN;
  } else {
    plmysql_down(host);
  }

  if (hostname)
    myfree(hostname);
}

static int plmysql_query(mysql_t * setup, char *query)
{
  sqlhost_t *host;
  int     i;
  int 	  toret;

  /* Try to connect to a mysql database */
  for (i = 0; i < (setup->plsql->len_hosts); i++) {
    host = &(setup->plsql->db_hosts[i]);

    switch(host->stat) {
      case STAT_OK:
        plmysql_connect(host, setup->sqlconf);
        break;

      case STAT_BAD:
        if(host->ts < time(NULL)) {
	  host->stat=STAT_OK;
          plmysql_connect(host, setup->sqlconf);
	}
        break;
    }

    if (host->stat == STAT_OPEN) {
      if (mysql_query(host->db, query)) 
        plmysql_down(host);
      else 
	break;
    }
  }

  /* if no connection succeded */
  if(host->stat != STAT_OPEN)
    return -1;

  /* Close connections left open */
  for (; i < (setup->plsql->len_hosts); i++) {
    host = &(setup->plsql->db_hosts[i]);
    if(host->stat == STAT_OPEN) {
      plmysql_close(host);
      host->stat = STAT_OK;
    }
  }

  return 0;
}

/*****
 * mysql_add */
int mysql_add(mysql_t * data, char *clientip, int time)
{
    char * escaped_string;
    static VSTRING *query;

    /* Prepare data */
    escaped_string = (char *)(mymalloc(sizeof(char)*(strlen(clientip)*2)+1));
    if(escaped_string == NULL)
      msg_fatal("optm_mysql, mysql_add: out of memory");

    mysql_escape_string(escaped_string, clientip, (unsigned int) strlen(clientip));

    /* prepare the query */
    query = vstring_alloc(24);
    vstring_sprintf(query, "insert into %s set %s = '%d', %s = '%s'", data->sqlconf->table, data->sqlconf->select_field, 
							    	time, data->sqlconf->where_field, clientip);
    myfree(escaped_string);

    if (plmysql_query(data, vstring_str(query))) { 
      vstring_free(query);
      return 1;
    }
    vstring_free(query);
    return 0;
}

/*****
 * mysql_removed */
int mysql_remove(mysql_t * data, int time)
{
    static VSTRING *query;
    int toret=0;

    /* prepare the query */
    query = vstring_alloc(24);
    vstring_sprintf(query, "delete from %s where %s < %d AND %s != 0", data->sqlconf->table, data->sqlconf->select_field, 
							    	time, data->sqlconf->select_field);
    if (plmysql_query(data, vstring_str(query)))
      toret=1;

    vstring_free(query);
    return toret;
}

/* plmysql_dealloc - free memory associated with PLMYSQL close databases */
static void plmysql_free(plsql_t *plsql)
{
    int     i;

    for (i = 0; i < plsql->len_hosts; i++) {
	if (plsql->db_hosts[i].db)
	  mysql_close(plsql->db_hosts[i].db);
	myfree(plsql->db_hosts[i].hostname);
    }
    myfree((char *) plsql->db_hosts);
    myfree((char *) plsql);
}

void mysql_localclose(mysql_t * setup) {
    int     i;

    plmysql_free(setup->plsql);
    myfree(setup->sqlconf->username);
    myfree(setup->sqlconf->password);
    myfree(setup->sqlconf->dbname);
    myfree(setup->sqlconf->table);
    myfree(setup->sqlconf->select_field);
    myfree(setup->sqlconf->where_field);

    for (i = 0; i < setup->sqlconf->len_hosts; i++)
	myfree(setup->sqlconf->hostnames[i]);
    
    myfree((char *) setup->sqlconf->hostnames);
    myfree((char *) setup->sqlconf);

    return;
}

////#ifdef HAS_MYSQL
struct optm_t optm_mysql = {
  "mysql",
  (void *(*)(char *))(mysql_open),
  (void (*)(void *))(mysql_localclose),
  (int (*)(void *, int))(mysql_remove),
  (int (*)(void *, char *, int))(mysql_add)
};
////#endif



#endif
#endif
